// Dan  5/25/2004  \DanLP110\Include\DataValue.h
#pragma once

#include "Allocate.h"

class DataValue final {
public:
    BYTE *m_data{nullptr};
    int m_length{-1};

    DataValue() = default;

    DataValue(const BYTE *data, const int length) {
        SetData(data, length);
    }

/* unused
    DataValue(const DataValue &other) {
        *this = other;
    }
*/
    ~DataValue() {
        FreeData();
    }

    void Init() {
        m_data = nullptr;
        m_length = 0;
    }

/* unused
    DataValue &operator=(const DataValue &other) {
        AllocateData(other.GetLength());
        if (m_data)
            memcpy(m_data, other.m_data, m_length);
        return *this;
    }
*/
    DataValue(const DataValue &) = delete;
    DataValue &operator=(const DataValue &) = delete;
    DataValue(DataValue &&) = delete;
    DataValue &operator=(DataValue &&) = delete;

    bool operator ==(const DataValue &other) const noexcept {
        if (GetLength() != other.GetLength())
            return false;

        if (m_length == 0)
            return true;

        return (memcmp(m_data, other.m_data, m_length) == 0);
    }

    [[nodiscard]] void *GetData() const {
        return m_data;
    }

    void SetData(const BYTE *data, const int length) {
        AllocateData(length);
        if (m_data)
            memcpy(m_data, data, m_length);
    }

    [[nodiscard]] int GetLength() const {
        return m_length;
    }

    void *AllocateData(const int length) {
        if (m_length == length)
            return m_data;

        if (m_data)
            FreeExt(m_data);

        m_length = 0;

        if (length)
            m_data = (BYTE *)AllocExt(length);
        else
            m_data = nullptr;

        if (m_data)
            m_length = length;

        return m_data;
    }

    void FreeData() {
        if (m_length) {
            FreeExt(m_data);
            m_length = 0;
            m_data = nullptr;
        }
    }

    [[nodiscard]] bool IsEmpty() const {
        return m_length == 0;
    }
};
