// ReSharper disable CppUnusedIncludeDirective
#pragma once

#include "StdAfx.h"

#ifndef _CSTDINT_
#include <cstdint>
#endif

#ifndef _LIMITS_
#include <limits>
#endif

#include "FileCons.h"
#pragma warning(push)
#pragma warning(disable: 4056)
namespace hash_constants {
    //const float smallest = 1.17549436e-38f;      // 0x00800000
    //const float greatest = 1.0842022e-19F;       // 0x20000000
    //const float squareRootOfSmallest = 3.4028234e38f;        // 0x7F7FFFFF
    //const float squareRootOfGreatest = 1.8446744e19F;        // 0x5F800000

    //    constexpr float positiveInfinity = 3.4028234e38f * 2.F;
    //    constexpr float negativeInfinity = 3.4028234e38f * -2.F;
    //    constexpr float NaN = std::numeric_limits<float>::quiet_NaN();// 3.4028234e38f * 2.F * 0.F;
    constexpr int big_int = 2147483647;
    constexpr unsigned int max_uint = 0xffffffff;

    constexpr float big_float = 1e8F;
    constexpr float minus_big_float = -1e8F;
    constexpr float smalltan = 1.0e-6F;
    constexpr float max_float = 3.402823466e+38F;
    constexpr float pi = 3.1415926535F; // pi
    constexpr float pi_2 = 3.1415926535F * 2.F; // pi multiplicate 2
    constexpr float pi_div2 = 3.1415926535F / 2.F; // pi divided 2
    constexpr float pi_div8 = 3.1415926535F / 8.F; // pi divided 8
    constexpr float sqrt2 = 1.4142135623731F; // square root of two
    constexpr float sqrt3 = 1.73205076F; // square root of three
    constexpr float reci_sqrt3 = 0.57735027F; // reciproke square root of three
    constexpr float e2 = 2.7182818F * -2.F; // -2e
    constexpr float gravity = 980.665F; // 980.665 CM/SEC/SEC
    constexpr float radians = 57.29578F;
    constexpr float small_float = 0.001F;
    //double constants
    constexpr double big_int_double = 2147483647.;
    constexpr double double_sqrt2 = 1.41421356237309504880; // square root of two
    constexpr double double_pi = 3.14159265358979323846; // pi
}

#if defined(__AVX2__) || defined(__AVX__) || defined(AVX_I)
#include <intrin.h>
#endif

/*
#ifdef __AVX__
#include <immintrin.h>
//#include <xmmintrin.h>     //SSE
//#include <emmintrin.h>     //SSE2
//#include <pmmintrin.h>     //SSE3
//#include <tmmintrin.h>     //SSSE3
//#include <smmintrin.h>     //SSE4.1
//#include <nmmintrin.h>     //SSE4.2
#else
#include "pmmintrin.h"
#endif
*/

template <typename T1>
bool IsFinite(T1 d) noexcept {
    return isfinite(d);
}

__forceinline bool IsFinite(const float d) noexcept {
#if defined(__INTEL_COMPILER) || defined(__INTEL_LLVM_COMPILER)
    return isfinitef(d);
#else
    return isfinite(d);
#endif
}

namespace hash_math {
    template <typename T>
    T abs(T a) noexcept {
        return a < 0 ? -a : a;
    }

    template <typename T>
    T sqr(const T x) noexcept {
        return x * x;
    }

    constexpr float sqr(const float x) noexcept {
        return x * x;
    }

#ifdef _DEBUG
    __forceinline float rcp(const float x) noexcept {
        ASSERT(IsFinite(x));
#else
    constexpr __forceinline float rcp(const float x) noexcept {
#endif
        return x != 0.F ? 1.F / x : 0.F;
    }

    __forceinline float sqrt(const float x) noexcept {
        ASSERT(IsFinite(x));
        return x > 0.F ? sqrtf(x) : 0.F;
    }

    __forceinline float abs(const float x) noexcept {
        return fabsf(x);
    }

    __forceinline float acos(const float x) noexcept {
        ASSERT(IsFinite(x));
        ASSERT(x <= 1.F && x >= -1.F);
        return acosf(x);
    }

    __forceinline float asin(const float x) noexcept {
        ASSERT(IsFinite(x));
        ASSERT(x <= 1.F && x >= -1.F);
        return asinf(x);
    }

    __forceinline float atan(const float x) noexcept {
        return atanf(x);
    }

    __forceinline float atan2(const float y, const float x) noexcept {
        return atan2f(y, x);
    }

    __forceinline float cos(const float x) noexcept {
        return cosf(x);
    }

    __forceinline float cosh(const float x) noexcept {
        return coshf(x);
    }

    __forceinline float exp(const float x) noexcept {
        return expf(x);
    }

    __forceinline float fmod(const float x, const float y) noexcept {
        return fmodf(x, y);
    }

    __forceinline float log(const float x) noexcept {
        return logf(x);
    }

    __forceinline float log10(const float x) noexcept {
        return log10f(x);
    }

    __forceinline float pow(const float x, const float y) noexcept {
        return powf(x, y);
    }

    __forceinline float sin(const float x) noexcept {
        return sinf(x);
    }

    __forceinline float sinh(const float x) noexcept {
        return sinhf(x);
    }

    __forceinline float tan(const float x) noexcept {
        return tanf(x);
    }

    __forceinline float tanh(const float x) noexcept {
        return tanhf(x);
    }

    __forceinline float floor(const float x) noexcept {
        return floorf(x);
    }

    __forceinline float ceil(const float x) noexcept {
        return ceilf(x);
    }

    __forceinline float frac(const float x) noexcept {
        return x - floor(x);
    }

    // version for double
    __forceinline double abs(const double x) noexcept {
        return fabs(x);
    }

    __forceinline double acos(const double x) noexcept {
        return ::acos(x);
    }

    __forceinline double asin(const double x) noexcept {
        return ::asin(x);
    }

    __forceinline double atan(const double x) noexcept {
        return ::atan(x);
    }

    __forceinline double atan2(const double y, const double x) noexcept {
        return ::atan2(y, x);
    }

    __forceinline double cos(const double x) noexcept {
        return ::cos(x);
    }

    __forceinline double cosh(const double x) noexcept {
        return ::cosh(x);
    }

    __forceinline double exp(const double x) noexcept {
        return ::exp(x);
    }

    __forceinline double fmod(const double x, const double y) noexcept {
        return ::fmod(x, y);
    }

    __forceinline double log(const double x) noexcept {
        return ::log(x);
    }

    __forceinline double log10(const double x) noexcept {
        return ::log10(x);
    }

    __forceinline double pow(const double x, const double y) noexcept {
        return ::pow(x, y);
    }

#ifdef _DEBUG
    __forceinline double rcp(const double x) noexcept {
        ASSERT(IsFinite(x));
#else
    constexpr __forceinline double rcp(const double x) noexcept {
#endif
        return x != 0. ? 1. / x : 0.;
    }

    __forceinline double sin(const double x) noexcept {
        return ::sin(x);
    }

    __forceinline double sinh(const double x) noexcept {
        return ::sinh(x);
    }

    constexpr double sqr(const double x) noexcept {
        return x * x;
    }

    __forceinline double sqrt(const double x) noexcept {
        ASSERT(IsFinite(x));
        return x > 0.F ? ::sqrt(x) : 0.F;
    }

    __forceinline double tan(const double x) noexcept {
        return ::tan(x);
    }

    __forceinline double tanh(const double x) noexcept {
        return ::tanh(x);
    }

    __forceinline double floor(const double x) noexcept {
        return ::floor(x);
    }

    __forceinline double ceil(const double x) noexcept {
        return ::ceil(x);
    }
}

//replacement for float ==
__forceinline bool floatEQ(const float f1, const float f2) noexcept {
#ifdef _DEBUG
    const __m128 compare1 = _mm_cmpeq_ss(_mm_set_ss(f1), _mm_set_ss(f2));
    bool t1 = (bool)_mm_movemask_ps(compare1);
    bool t2 = f1 == f2;
    ASSERT(t1==t2);
#endif
    const __m128 compare = _mm_cmpeq_ss(_mm_set_ss(f1), _mm_set_ss(f2));
    return (bool)_mm_movemask_ps(compare);
}

//replacement for float !=
__forceinline bool floatNEQ(const float f1, const float f2) noexcept {
#ifdef _DEBUG
    const __m128 compare1 = _mm_cmpneq_ss(_mm_set_ss(f1), _mm_set_ss(f2));
    bool t1 = (bool)_mm_movemask_ps(compare1);
    bool t2 = f1 != f2;
    ASSERT(t1==t2);
#endif
    const __m128 compare = _mm_cmpneq_ss(_mm_set_ss(f1), _mm_set_ss(f2));
    return (bool)_mm_movemask_ps(compare);
}

constexpr float fast_ftol_signedf(const float d) noexcept {
    return static_cast<float>(static_cast<int>(d));
}

__forceinline float fast_ftol_unsignedf(const float d) noexcept {
    return fabsf(truncf(d));
}

template <typename T1>
int fast_ftol_signed(T1 d) noexcept {
    return static_cast<int>(d);
}

__forceinline int fast_ftol_signed(const float d) noexcept {
    return _mm_cvtt_ss2si(_mm_set_ss(d));
}

__forceinline int fast_ftol_signed(const double d) noexcept {
    return _mm_cvttsd_si32(_mm_set_sd(d));
}

template <typename T1>
int fast_ftol_unsigned(T1 d) noexcept {
    return abs(static_cast<int>((float)d));
}

__forceinline int fast_ftol_unsigned(const float d) noexcept {
    return hash_math::abs(_mm_cvtt_ss2si(_mm_set_ss(d)));
}

__forceinline int fast_ftol_unsigned(const double d) noexcept {
    return hash_math::abs(_mm_cvttsd_si32(_mm_set_sd(d)));
}

template <typename T1>
int64_t fast_ftol_unsigned_longlong(const T1 d) noexcept {
    //for time::ticks
    return hash_math::abs(static_cast<int64_t>(d));
}

// faster than normal ::floor/::floorf
//rounding down to int (example 2.9 -> 2)
template <typename T1>
int fast_floor(T1 d) noexcept {
    return static_cast<int>(hash_math::floor(d));
}

#if defined(__AVX2__) || defined(__AVX__) || defined(AVX_I)
__forceinline int fast_floor(const float d) noexcept {
    return _mm_cvtt_ss2si(_mm_round_ps(_mm_load_ss(&d), (_MM_FROUND_TO_NEG_INF | _MM_FROUND_NO_EXC)));
}

__forceinline int fast_floor(const double d) noexcept {
    return _mm_cvttsd_si32(_mm_round_pd(_mm_load_sd(&d), (_MM_FROUND_TO_NEG_INF | _MM_FROUND_NO_EXC)));
}
#endif

//rounding to the nearest int (example 2.9 -> 3)
template <typename T1>
int fast_round(T1 d) noexcept {
    if (d < 0) {
        return static_cast<int>(d - 0.5);
    }
    return static_cast<int>(d + 0.5); // NOLINT(bugprone-incorrect-roundings)
}

#if defined(__AVX2__) || defined(__AVX__) || defined(AVX_I)
__forceinline int fast_round(const float d) noexcept {
    return _mm_cvtt_ss2si(_mm_round_ps(_mm_load_ss(&d), (_MM_FROUND_TO_NEAREST_INT | _MM_FROUND_NO_EXC)));
}

__forceinline int fast_round(const double d) noexcept {
    return _mm_cvttsd_si32(_mm_round_pd(_mm_load_sd(&d), (_MM_FROUND_TO_NEAREST_INT | _MM_FROUND_NO_EXC)));
}
#endif

#pragma warning(pop)
