//YP  28/01/2005  \Yves120\Include\ImageIO.h
// ReSharper disable CppClangTidyCppcoreguidelinesProTypeMemberInit
#pragma once

#include "SDK/HBitmap.h"

#include "DllPlugi.h"

constexpr int IMAGEIODLLVERSION  = 1;

class PLUGINEXPORT HOutputInfo final {
public:
    int GetTime();
    int GetStartTime();
    int GetEndTime();
    int GetStep();
    int GetFPS();
    bool GetIsAudioStereo();
    int GetAudioFrequency();
    bool GetIsAlpha(); //User requested that the alpha channel be saved in the file
    int GetGammaPreset(); //One of the Gamma presets, the most characteristic being sRGB (or NTSC)
    float GetOutputGamma(); //Gamma value for image data correction prior to storing to file.
    float GetViewGamma(); //Gamma value of the user display setting
    String GetUser(); //A string containing the user name to store as metadata in output files.
    String GetOrganization(); //A string containing the user organisation (company, institution, etc).
    String GetURL(); //A string containing the user URL (web page).
    String GetEmail(); //A string containing the user email address.
    String GetApplicationID(); //A string containing "Hash Animation:Master vn" where vn is the A:M version number.
};

class ImageIOInfo final {
public:
    String m_openfilter; //OpenFile dialog file filter. The format is "|Display string|filter string|Macintosh filetype||"
    String m_savefilter; //Save dialog file filter. The format is "|Display string|filter string|Macintosh filetype||"
    String m_suffix; //The prefered or most common file extension. Do not include the dot.
    String m_matchname; //A unique name used to identify the plugin. This name must be unique among plugins and does not contain spaces
    String m_localizedname; //The name of the supported file type as displayed in the Render Option drop down list.
    String m_documentation; //A HTML string describing the plugin and its usage.
// @formatter:off 
    union {
        BITFIELDS8 m_ImageIOInfo_flags{0};

        struct {
            //This plugin can :
            BITFIELDS8 m_candomultiframes : 1, //- process (read-write) multiple frames in one single file. QuickTime and AVI can do that. But Gif can too.
                m_canwrite : 1, //- write to file. All plugins must have read from file capability but may not write to file
                m_haswriteoptions : 1, //- Provides a callable Save Options dialog box. Most common are compression settings
                m_canstorealpha : 1, //- store alpha channels with the image data
                m_canstoresound : 1, //- store sound data. Either in same file as images or on separate file.
                m_canstoremultibuffers : 1, //- store multiple buffers. OpenEXR and Jpeg2000 can do that.
                m_closewrite : 1; //- needs to close animation file after saving frames
        };
    };
// @formatter:on

    ImageIOInfo() = default;
};

//NOTE: A GammaType of GAMMA_NTSC is equivalent to 2.2 but ideally should follow the CIE Rec 709 with a slope of 4.5 below 0.018 and power 0.45 above 0.018
enum GammaType {
    GAMMA_NONE,
    GAMMA_MONITOR,
    GAMMA_NTSC,
    GAMMA_CRT,
    GAMMA_CUSTOM
};

// ReSharper disable once CppClassCanBeFinal
class ImageIO : public DLLPlugin {
    //don't final , iop lugins has a derivat
public:
    bool m_isopened{false};

    ImageIO() = default;
    explicit ImageIO(HTreeObject *htreeobject) : DLLPlugin(htreeobject) {}

    //ImageIO users should call this write and closewrite when done
    __forceinline long Write(HBitmapNode *bn, HOutputInfo *outputinfo, const String &filename, short *wavedata, bool hasaudio);
    __forceinline void CloseWrite(bool hasaudio);

    //ImageIO developers should write these
    //virtual bool Load( BitmapNode &bn, const String &filename ) { return false; }
    //virtual char *GetHandleFileType() { return nullptr; }
    virtual bool CanLoadImage(const String &filename) {
        return false;
    }

    virtual void GetFilesStartEndFrame(const String &filename, int &start, int &end) {
        start = -1;
        end = -2;
    }

    virtual bool SetSaveOptions(CWnd *parent, const String &extension, float fps) {
        return false;
    }

    virtual bool GetImageIOInfo(ImageIOInfo &iioinfo) {
        return false;
    }

    virtual long OpenWriteFrame(String filename, HBitmapNode *bn, HOutputInfo *outputinfo, bool hasaudio) {  // NOLINT(performance-unnecessary-value-param)
        return 0;
    }

    virtual long WriteFrame(HBitmapNode *bn, HOutputInfo *outputinfo, const String &filename, short *wavedata, bool hasaudio) {
        return HBitmapNode::WRITE_NORESULT;
    }

    virtual void CloseWriteFrame(bool hasaudio) {}

    virtual long LoadFrame(HBitmapNode *bn, const String &filename, int frame = -1, UINT flags = 0) {
        return HBitmapNode::LOAD_FAILED;
    }

    virtual bool IsFileOpen() {
        return false;
    }

};

void ImageIO::CloseWrite(const bool hasaudio) {
    if (m_isopened)
        CloseWriteFrame(hasaudio);
    m_isopened = false;
}

long ImageIO::Write(HBitmapNode *bn, HOutputInfo *outputinfo, const String &filename, short *wavedata, const bool hasaudio) {
    try {
        if (!m_isopened) {
            if (const long rvalue = OpenWriteFrame(filename, bn, outputinfo, hasaudio); rvalue != 0)
                if (rvalue == HBitmapNode::WRITE_ERROR || rvalue < 0) //negativ values from avi are errors too
                    return HBitmapNode::WRITE_ERROR;
            //return rvalue;
            m_isopened = true;
        }
        return WriteFrame(bn, outputinfo, filename, wavedata, hasaudio);
    }
    catch (...) {
        CString err;
        err.Format("Unhandled exception in the \"%s\" image io plugin.", GetClassName());
        AfxMessageBox(err, MB_OK);
        return HBitmapNode::WRITE_ERROR;
    }
}

