// ReSharper disable CppNonExplicitConvertingConstructor
// ReSharper disable CppRedundantParentheses
// ReSharper disable CppClangTidyCppcoreguidelinesProTypeMemberInit
#pragma once

#include "Vector.h"

#if defined(__AVX2__) || defined(__AVX__) || defined(AVX_I)
#include <intrin.h>
const __m128 zero_m128 = _mm_set_ps(1.F, 0.F, 0.F, 0.F);//_mm_setzero_ps();
#else
#define vec_splat(x, e) _mm_shuffle_ps(x, x, _MM_SHUFFLE(e, e, e, e))

static __forceinline __m128 dot3(const __m128 vec0, const __m128 vec1) noexcept {
    const __m128 result = _mm_mul_ps(vec0, vec1);
    return _mm_add_ps(vec_splat(result, 0), _mm_add_ps(vec_splat(result, 1), vec_splat(result, 2)));
}
#endif
#ifndef __SMMINTRIN_H
#include <smmintrin.h>
#endif

/*
! don't use !
    precision only 11 bit
    [[nodiscard]] float InvNorm() const {
        return _mm_cvtss_f32(_mm_rsqrt_ss(_mm_dp_ps(mmvalue, mmvalue, 0x7F))); //SSE4
        return _mm_cvtss_f32(_mm_rsqrt_ss(dot3(mmvalue, mmvalue)));
    }

//too big error -> differences between the versions
static __forceinline __m128 newtonrapson_rsqrt4(const __m128 v) {
    const __m128 approx = _mm_rsqrt_ps(v);
    const __m128 muls = _mm_mul_ps(_mm_mul_ps(v, approx), approx);
    return _mm_mul_ps(_mm_mul_ps(_mm_set_ps(.5f, .5f, .5f, .5f), approx), _mm_sub_ps(_mm_set_ps(3.F, 3.F, 3.F, 3.F), muls));
}
//relative error bigger than above
__forceinline __m128 newtonrapson_rsqrt(const __m128& a) {
    const __m128 r = _mm_rsqrt_ps(a);
    return _mm_add_ps(_mm_mul_ps(_mm_set_ps1(1.5f), r), _mm_mul_ps(_mm_mul_ps(_mm_mul_ps(a, _mm_set_ps1(-0.5F)), r), _mm_mul_ps(r, r)));
}
*/

class alignas(16) Vector3SSE final {
public:
    //Member variables
    union {
        __m128 mmvalue{0.F, 0.F, 0.F, 1.F};
        struct {
            float x, y, z;
            int dummy;
        };
    };

    //constructors
    constexpr Vector3SSE() noexcept = default;

    constexpr explicit Vector3SSE(const __m128 m) noexcept : mmvalue(m) {}
    explicit Vector3SSE(const Vector &other) noexcept : mmvalue(_mm_set_ps(1.F, other.z, other.y, other.x)) {}
    explicit Vector3SSE(const float fill) noexcept : mmvalue(_mm_set_ps(1.F, fill, fill, fill)) {}
    Vector3SSE(const float xi, const float yi, const float zi) noexcept : mmvalue(_mm_set_ps(1.F, zi, yi, xi)) {}

    void Set(const float s_x, const float s_y, const float s_z) noexcept {
        mmvalue = _mm_set_ps(1.F, s_z, s_y, s_x);
    }

    void Set(const float fill) noexcept {
        mmvalue = _mm_set1_ps(fill);
    }

    Vector3SSE &operator=(const Vector other) noexcept {
        mmvalue = _mm_set_ps(1.F, other.z, other.y, other.x);
        return *this;
    }

    [[nodiscard]] constexpr Vector ToVector() const noexcept {
        return Vector(x, y, z);
    }

    explicit constexpr operator const __m128 &() const noexcept {
        return mmvalue;
    }

    explicit constexpr operator __m128 &() noexcept {
        return mmvalue;
    }

    static Vector3SSE load(const void *const a) noexcept {
        return Vector3SSE(_mm_and_ps(_mm_load_ps((float *)a), _mm_castsi128_ps(_mm_set_epi32(0, -1, -1, -1))));
    }

    static Vector3SSE loadu(const void *const a) noexcept {
        return Vector3SSE(_mm_and_ps(_mm_loadu_ps((float *)a), _mm_castsi128_ps(_mm_set_epi32(0, -1, -1, -1))));
    }

    static void storeu(void *ptr, const Vector3SSE &v) noexcept {
        _mm_storeu_ps((float *)ptr, v.mmvalue);
    }

    constexpr float const &operator [](const int index) const noexcept {
        return (&x)[index];
    }

    constexpr float &operator [](const int index) noexcept {
        return (&x)[index];
    }

    bool operator ==(const Vector3SSE &b) const noexcept {
        return _mm_movemask_ps(_mm_cmpeq_ps(mmvalue, b.mmvalue)) == 0xf;
    }

    bool operator !=(const Vector3SSE &b) const noexcept {
        return _mm_movemask_ps(_mm_cmpneq_ps(mmvalue, b.mmvalue)) != 0;
    }

    //arithmetic operators with Vector3SSE
    Vector3SSE operator +(const Vector3SSE b) const noexcept {
        return Vector3SSE(_mm_add_ps(mmvalue, b.mmvalue));
    }

    Vector3SSE operator -(const Vector3SSE b) const noexcept {
        return Vector3SSE(_mm_sub_ps(mmvalue, b.mmvalue));
    }

    Vector3SSE operator *(const Vector3SSE b) const noexcept {
        return Vector3SSE(_mm_mul_ps(mmvalue, b.mmvalue));
    }

    Vector3SSE operator /(const Vector3SSE b) const noexcept {
        return Vector3SSE(_mm_div_ps(mmvalue, b.mmvalue));
    }

    Vector3SSE &operator +=(const Vector3SSE b) noexcept {
        mmvalue = _mm_add_ps(mmvalue, b.mmvalue);
        return *this;
    }

    Vector3SSE &operator +=(const Vector b) noexcept {
        const Vector3SSE local(b);
        mmvalue = _mm_add_ps(mmvalue, local.mmvalue);
        return *this;
    }

    Vector3SSE &operator -=(const Vector3SSE b) noexcept {
        mmvalue = _mm_sub_ps(mmvalue, b.mmvalue);
        return *this;
    }

    Vector3SSE &operator -=(const Vector b) noexcept {
        const Vector3SSE local(b);
        mmvalue = _mm_sub_ps(mmvalue, local.mmvalue);
        return *this;
    }

    Vector3SSE &operator *=(const Vector3SSE b) noexcept {
        mmvalue = _mm_mul_ps(mmvalue, b.mmvalue);
        return *this;
    }

    Vector3SSE &operator *=(const Vector b) noexcept {
        const Vector3SSE local(b);
        mmvalue = _mm_mul_ps(mmvalue, local.mmvalue);
        return *this;
    }

    Vector3SSE &operator /=(const Vector3SSE b) noexcept {
        mmvalue = _mm_div_ps(mmvalue, b.mmvalue);
        return *this;
    }

    Vector3SSE &operator /=(const Vector b) noexcept {
        const Vector3SSE local(b);
        mmvalue = _mm_div_ps(mmvalue, local.mmvalue);
        return *this;
    }

    //arithmetic operators with float
    Vector3SSE operator +(const float b) const noexcept {
        return Vector3SSE(_mm_add_ps(mmvalue, _mm_set1_ps(b)));
    }

    Vector3SSE operator -(const float b) const noexcept {
        return Vector3SSE(_mm_sub_ps(mmvalue, _mm_set1_ps(b)));
    }

    Vector3SSE operator *(const float b) const noexcept {
        return Vector3SSE(_mm_mul_ps(mmvalue, _mm_set1_ps(b)));
    }

    Vector3SSE operator /(const float b) const noexcept {
        if (b != 0.F)
            return Vector3SSE(_mm_div_ps(mmvalue, _mm_set1_ps(b)));
        return Vector3SSE(_mm_set_ps1(0.F));
    }

    //op= operators with float
    Vector3SSE &operator +=(const float b) noexcept {
        mmvalue = _mm_add_ps(mmvalue, _mm_set1_ps(b));
        return *this;
    }

    Vector3SSE &operator -=(const float b) noexcept {
        mmvalue = _mm_sub_ps(mmvalue, _mm_set1_ps(b));
        return *this;
    }

    Vector3SSE &operator *=(const float b) noexcept {
        mmvalue = _mm_mul_ps(mmvalue, _mm_set1_ps(b));
        return *this;
    }

    Vector3SSE &operator /=(const float b) noexcept {
        if (b != 0.F)
            mmvalue = _mm_div_ps(mmvalue, _mm_set1_ps(b));
        return *this;
    }

    //cross product NO constexpr
    [[nodiscard]] Vector3SSE Cross(const Vector3SSE b) const noexcept {
        return Vector3SSE(
            _mm_sub_ps(
                _mm_mul_ps(_mm_shuffle_ps(mmvalue, mmvalue, _MM_SHUFFLE(3, 0, 2, 1)), _mm_shuffle_ps(b.mmvalue, b.mmvalue, _MM_SHUFFLE(3, 1, 0, 2))),
                _mm_mul_ps(_mm_shuffle_ps(mmvalue, mmvalue, _MM_SHUFFLE(3, 1, 0, 2)), _mm_shuffle_ps(b.mmvalue, b.mmvalue, _MM_SHUFFLE(3, 0, 2, 1)))));
    }

    void CrossSelf(const Vector3SSE b) noexcept {
        mmvalue = _mm_sub_ps(
            _mm_mul_ps(_mm_shuffle_ps(mmvalue, mmvalue, _MM_SHUFFLE(3, 0, 2, 1)), _mm_shuffle_ps(b.mmvalue, b.mmvalue, _MM_SHUFFLE(3, 1, 0, 2))),
            _mm_mul_ps(_mm_shuffle_ps(mmvalue, mmvalue, _MM_SHUFFLE(3, 1, 0, 2)), _mm_shuffle_ps(b.mmvalue, b.mmvalue, _MM_SHUFFLE(3, 0, 2, 1))));
    }

    void CrossSelf(const Vector b) noexcept {
        const Vector3SSE local(b);
        mmvalue = _mm_sub_ps(
            _mm_mul_ps(_mm_shuffle_ps(mmvalue, mmvalue, _MM_SHUFFLE(3, 0, 2, 1)), _mm_shuffle_ps(local.mmvalue, local.mmvalue, _MM_SHUFFLE(3, 1, 0, 2))),
            _mm_mul_ps(_mm_shuffle_ps(mmvalue, mmvalue, _MM_SHUFFLE(3, 1, 0, 2)), _mm_shuffle_ps(local.mmvalue, local.mmvalue, _MM_SHUFFLE(3, 0, 2, 1))));
    }

    Vector3SSE operator ^(const Vector3SSE a) const noexcept {
        return Cross(a);
    } //same as Cross

    [[nodiscard]] Vector3SSE Negate() const noexcept {
        constexpr __m128 signs = {-0.F, -0.F, -0.F, -0.F};
        return Vector3SSE(_mm_xor_ps(mmvalue, signs));
    }

    Vector3SSE operator -() const noexcept {
        constexpr __m128 signs = {-0.F, -0.F, -0.F, -0.F};
        return Vector3SSE(_mm_xor_ps(mmvalue, signs));
    } //unary minus

    void FixNormalizedVector() noexcept {
        //fix when Normalize returns a Vector like this (0,1.000,1e-6)
        const Vector3SSE temp = Absolute();
        if (const bool x_one = temp.x > .99999f && temp.y < 1e-6F && temp.z < 1e-6F; x_one) {
            Set(SIGN_FLOAT(x), 0.F, 0.F);
            return;
        }
        if (const bool y_one = temp.y > .99999f && temp.x < 1e-6F && temp.z < 1e-6F; y_one) {
            Set(0.F, SIGN_FLOAT(y), 0.F);
            return;
        }
        if (const bool z_one = temp.z > .99999f && temp.x < 1e-6F && temp.y < 1e-6F; z_one) {
            Set(0.F, 0.F, SIGN_FLOAT(z));
        }
    }

#if defined(__AVX2__) || defined(__AVX__) || defined(AVX_I)
    [[nodiscard]] float Dot(const Vector3SSE b) const noexcept {
        return _mm_cvtss_f32(_mm_dp_ps(mmvalue, b.mmvalue, 0x7F));
    }

    [[nodiscard]] float Dot() const noexcept {
        return _mm_cvtss_f32(_mm_dp_ps(mmvalue, mmvalue, 0x7F));
    }

    [[nodiscard]] float Norm() const noexcept {
        return _mm_cvtss_f32(_mm_sqrt_ss(_mm_dp_ps(mmvalue, mmvalue, 0x7F)));
    }

    [[nodiscard]] bool IsZeroVector() noexcept {
#ifdef __clang__
        mmvalue[3] = 1.F;
#else
        mmvalue.m128_f32[3] = 1.F;
#endif
        return _mm_movemask_ps(_mm_cmpeq_ps(mmvalue, zero_m128)) == 0xf;
    }

    Vector3SSE &FixedNormalize() noexcept {
        if (IsZeroVector())
            return *this;
        const float square = Dot();
        if (!square) {
            Set(0.F);
            return *this;
        }
        *this *= hash_math::rcp(hash_math::sqrt(square));
        FixNormalizedVector();
        return *this;
    }

    Vector FixedNormalizeToVector() noexcept {
        if (IsZeroVector())
            return Vector();
        const float square = Dot();
        if (!square) {
            return Vector();
        }
        *this *= hash_math::rcp(hash_math::sqrt(square));
        FixNormalizedVector();
        return Vector(x, y, z);
    }

    Vector3SSE &Normalize() noexcept {
        if (IsZeroVector())
            return *this;
        const float square = Dot();
        return square != 0.F ? *this *= hash_math::rcp(hash_math::sqrt(square)) : *this;
    }

    float NormalizeLength() noexcept {
        if (IsZeroVector())
            return 0.F;
        const float length_local = Norm();
        if (length_local == 0.F)
            return 0.F;

        *this *= hash_math::rcp(length_local);
        return length_local;
    }
#else
    [[nodiscard]] float Dot(const Vector3SSE b) const noexcept {
        return _mm_cvtss_f32(dot3(mmvalue, b.mmvalue));
    }

    [[nodiscard]] float Dot() const noexcept {
        return _mm_cvtss_f32(dot3(mmvalue, mmvalue));
    }

    [[nodiscard]] float Norm() const noexcept {
        return _mm_cvtss_f32(_mm_sqrt_ss(dot3(mmvalue, mmvalue)));
    }

    Vector3SSE &Normalize() noexcept {
        const float square = Dot();
        return square != 0.F ? *this *= hash_math::rcp(hash_math::sqrt(square)) : *this;
    }

    Vector3SSE &FixedNormalize() noexcept {
        if (const float square = Dot(); square) {
            *this *= hash_math::rcp(hash_math::sqrt(square));
            FixNormalizedVector();
        }
        return *this;
    }

    Vector FixedNormalizeToVector() noexcept {
        const float square = Dot();
        if (!square) {
            return Vector();
        }
        *this *= hash_math::rcp(hash_math::sqrt(square));
        FixNormalizedVector();
        return Vector(x, y, z);
    }

    float NormalizeLength() noexcept {
        const float length = Norm();
        if (length != 0.F) {
            *this *= hash_math::rcp(length);
        }
        return length;
    }
#endif

    friend void swap(Vector3SSE &a, Vector3SSE &b) noexcept {
        const Vector3SSE tmp(a);
        a = b;
        b = tmp;
    }

    friend float Dot(const Vector3SSE a, const Vector3SSE b) noexcept {
        return a.Dot(b);
    }

    friend Vector3SSE Cross(const Vector3SSE a, const Vector3SSE b) noexcept {
        return a.Cross(b);
    }

    [[nodiscard]] Vector3SSE Absolute() const noexcept {
        const __m128 mask = _mm_castsi128_ps(_mm_set1_epi32(0x7fffffff));
        return Vector3SSE(_mm_and_ps(mmvalue, mask));
    }

    [[nodiscard]] bool IsLess(const float minvalue) const noexcept {
        const __m128 mask = _mm_castsi128_ps(_mm_set1_epi32(0x7fffffff));
        const auto a = Vector3SSE(_mm_and_ps(mmvalue, mask));
        if (a.x > minvalue)
            return false;
        if (a.y > minvalue)
            return false;
        if (a.z > minvalue)
            return false;
        return true;
    }

    void PushMinMax(Vector3SSE &minv, Vector3SSE &maxv) const noexcept {
        minv.mmvalue = _mm_min_ps(mmvalue, minv.mmvalue);
        maxv.mmvalue = _mm_max_ps(mmvalue, maxv.mmvalue);
    }

    constexpr friend Vector3SSE Min(const Vector3SSE left, const Vector3SSE right) noexcept {
        return Vector3SSE(_mm_min_ps(left.mmvalue, right.mmvalue));
    }

    constexpr friend Vector3SSE Max(const Vector3SSE left, const Vector3SSE right) noexcept {
        return Vector3SSE(_mm_max_ps(left.mmvalue, right.mmvalue));
    }

#ifdef _DEBUG
    bool IsValid() const {  // NOLINT(modernize-use-nodiscard)
        if (!IsFinite(x))
            return false;
        if (!IsFinite(y))
            return false;
        if (!IsFinite(z))
            return false;
        return true;
    }
#endif
};

__forceinline Vector3SSE operator +(const float a, const Vector3SSE b) noexcept {
    return b + a;
}

__forceinline Vector3SSE operator -(const float a, const Vector3SSE b) noexcept {
    return Vector3SSE(_mm_set1_ps(a)) - b;
}

__forceinline Vector3SSE operator *(const float a, const Vector3SSE b) noexcept {
    return b * a;
}

__forceinline Vector3SSE operator /(const float a, const Vector3SSE b) noexcept {
    return Vector3SSE(_mm_set1_ps(a)) / b;
}

__forceinline void PushMinMax(const Vector p, Vector3SSE &minv, Vector3SSE &maxv) noexcept {
    minv.mmvalue = _mm_min_ps(Vector3SSE(p).mmvalue, minv.mmvalue);
    maxv.mmvalue = _mm_max_ps(Vector3SSE(p).mmvalue, maxv.mmvalue);
}
