// KB  5/13/2003  \Ken105\Plugin\Textures\Wave\Evaluate.cpp

#include "StdAfx.h"
#include "Custom.h"

HProperty *CustomTexture::GetPropertyAt(const int i) {
    switch (i) {
        case 0:
            return m_amplitude;
        case 1:
            return m_wavelength;
        case 2:
            return m_damping;
        case 3:
            return m_propagate;
        case 4:
            return m_cycle;
        default:
            return nullptr;
    }
}

bool CustomTexture::Evaluate(const Vector &evalp, HTexInfo *texinfo) {

    if (m_wavelength->GetValue() == 0.f)
        return TRUE;

    float riprad= evalp.Norm();
    if (riprad == 0.f)
        return TRUE;

    Vector ripple = evalp;
    ripple /= riprad;

    float damp, cycle = 0.f;
    if (m_cycle->GetValue())
        cycle = 2.f * fmod(texinfo->GetTime().GetSeconds(), m_cycle->GetValue()) / m_cycle->GetValue();
    riprad /= m_wavelength->GetValue() * 0.5f - m_propagate->GetValue() - cycle;

    if (m_damping->GetValue() == 100.f)
        damp = 1.f;
    else
        damp = hash_math::pow((double)m_damping->GetValue() / 100.0, (double)riprad);

    int iriprad;
    if (ISODD(iriprad = (int)riprad))
        ripple.Negate();
    if ((riprad = riprad - iriprad) > 0.5f)
        riprad = 1.f - riprad;
    riprad *= damp * m_amplitude->GetValue() * 0.5f;
    ripple *= riprad;

    *texinfo->GetN() += ripple;
    texinfo->GetN()->Normalize();

    return TRUE;
}
