// KB  9/26/01  \Ken90\Plugin\Turb\GridTurb\Evaluate.cpp

// ReSharper disable CppTooWideScope
#include "Custom.h"
#include "texfuncs.h"
#include "TurbApp.h"

extern CTurbApp theApp;

HProperty *CustomTurbulence::GetPropertyAt(const int i) {
    switch (i) {
        case 0:
            return m_celltype;
        case 1:
            return m_octaves;
        case 2:
            return m_gridsize;
        default:
            return nullptr;
    }
}

float CustomTurbulence::Evaluate(const Vector &evalp) {
    const Vector P = evalp;
    float turb;
    const float size = 1.f - m_gridsize->GetValue() / 100.f;
    float x = P.x;
    float y = P.y;
    const float z = P.z;
    float v, scale;
    const float r = m_octaves->GetValue() / 100.f;
    switch (m_celltype->GetValue()) {
        case 0:
            v = repeat(x,r);
            scale = pulse(size,1.f,0.1f,v);
            turb = scale;
            v = repeat(y,r);
            scale = pulse(size,1.f,0.1f,v);
            turb += scale;
            v = repeat(z,r);
            scale = pulse(size,1.f,0.1f,v);
            turb += scale;
            break;
        case 1:
            v = repeat(x,r);
            scale = pulse(size,1.f,0.1f,v);
            turb = scale;
            v = repeat(y,r);
            scale = pulse(size,1.f,0.1f,v);
            turb += scale;
            break;
        case 2:
            CalcUV(P, &x, &y, 0);
            v = repeat(x,r*100.f);
            scale = pulse(size,1.f,0.1f,v);
            turb = scale;
            v = repeat(y,r*100.f);
            scale = pulse(size,1.f,0.1f,v);
            turb += scale;
            break;
        case 3:

            CalcUV(P, &x, &y, 1);
            v = repeat(x,r*100.f);
            scale = pulse(size,1.f,0.1f,v);
            turb = scale;
            v = repeat(y,r*100.f);
            scale = pulse(size,1.f,0.1f,v);
            turb += scale;
            break;

        case 4:
            turb = EvaluateHex(P, size, r, 0.1f);
            break;
        default:
            turb = 0;
    }
    turb = turb > 1.f ? 1.f : turb;
    return turb;

}

void CustomTurbulence::CalcUV(const Vector e, float *u, float *v, const int type) const {
    float ele = 0;
    Vector temp;
    temp.x = e.x;
    temp.y = e.y;
    temp.z = e.z;
    temp.Normalize();
    const float az = (-hash_math::atan2(temp.y, temp.x) + hash_constants::pi) / (2.f * hash_constants::pi);
    if (type == 0) // spherical
        ele = 0.5f - hash_math::cos(temp.z) / hash_constants::pi;
    else if (type == 1) // cylindrical
        ele = temp.z;
    *u = az;
    *v = ele;

}

float CustomTurbulence::EvaluateHex(const Vector P, const float radius, float scale, float sfuzz) const {

    constexpr float tileradius = .2f;
    const float mortarwidth = radius / 10.f;
    const Vector p = P;
    constexpr float tilewidth = tileradius * 1.7320508f; /* sqrt(3) */
    float ss, s, t;

    CalcUV(p, &s, &t, 0);
    sfuzz *= 1 - t;
    s *= 2.42f;
    t *= 2.42f;

    sfuzz /= 2.42f;
    scale *= 100;
    s = mod(s * scale, 2.42f);
    t = mod(t * scale, 2.42f);
    const float tt = hash_math::fmod(t, 1.5f * tileradius);
    float ttile = (float)fast_floor(t / (1.5f * tileradius));
    if (fmod(ttile / 2, 1) == 0.5f)
        ss = s + tilewidth / 2.f;
    else
        ss = s;
    float stile = (float)fast_floor(ss / tilewidth);
    ss = hash_math::fmod(ss, tilewidth);
    float mortar;
    const float mw2 = mortarwidth / 2.f;
    if (tt < tileradius) {
        mortar = 1 - smoothstep(mw2, mw2 + sfuzz, ss) *
        (1 - smoothstep(tilewidth - mw2 - sfuzz, tilewidth - mw2, ss));
    } else {
        const float x = tilewidth / 2.f - hash_math::abs(ss - tilewidth / 2.f);
        const float y = 1.7320508f * (tt - tileradius);
        if (y > x) {
            if (fmod(ttile / 2, 1) == 0.5f)
                stile -= 1.f;
            ttile += 1.f;
            if (ss > tilewidth / 2.f)
                stile += 1.f;
        }
        mortar = smoothstep(x - 1.73f * mw2 - sfuzz, x - 1.73f * mw2, y) *
        (1 - smoothstep(x + 1.73f * mw2, x + 1.73f * mw2 + sfuzz, y));
    }

    [[maybe_unused]] float tileindex = stile + 41 * ttile;
    return mortar;
}

float CustomTurbulence::Voronoi2(const Vector *p, const int style) const {
    Vector rpos;
#ifdef _DEBUG
    Vector rpos2;
#endif
    float f1 = 1000.0f;
    float f2 = 1000.0f;
    const Vector thiscell((fast_floor(p->x)), (fast_floor(p->y)), (fast_floor(p->z)));
    for (int i = -1; i <= 1; ++i) {
        for (int j = -1; j <= 1; ++j) {
            for (int k = -1; k <= 1; ++k) {
                Vector add;
                Vector pos;
                add.x = (float)i;
                add.y = (float)j;
                add.z = (float)k;

                Vector testcell = thiscell + add;

                pos.x = testcell.x + theApp.CellNoise(testcell);
                pos.y = testcell.y + theApp.Noise(-testcell);
                pos.z = testcell.z + theApp.CellNoise(testcell + 2);
                Vector offset = pos - *p;
                const float dist = offset.Dot(offset);
                if (dist < f1) {
                    f2 = f1;
                    f1 = dist;
#ifdef _DEBUG
                    rpos2 = rpos;
#endif
                    rpos = pos;
                } else if (dist < f2) {
                    f2 = dist;
#ifdef _DEBUG
                    rpos2 = pos;
#endif
                }
            }
        }
    }
    f1 = hash_math::sqrt(f1);

    switch (style) {
        case 1:
            return f1;
        case 2:
            return theApp.Noise(rpos);
        case 3:
            f2 = hash_math::sqrt(f2);
            return f2 - f1;
        case 4:
            f2 = hash_math::sqrt(f2);
            return hash_math::sin((f2 - f1) * hash_constants::pi);
        case 5:
            f2 = hash_math::sqrt(f2);
            return f2;
        case 6:
            return hash_math::sin(f1 * (hash_constants::pi / 2.f));
        case 7:
        {
            f2 = hash_math::sqrt(f2);
#ifdef _DEBUG
            const Vector testcell = rpos2 - rpos;
            const float dt = hash_math::sqrt(testcell.Dot(testcell));
            [[maybe_unused]] float sf = dt / (f1 + f2);
#endif
            return (float)step(m_octaves->GetValue()/100.f,f2-f1);
        }
        default:
            return f1;
    }
}
