// RC  6/4/2002  \Randy95\Plugin\Turb\GridTurb\TurbApp.cpp
//

#include "StdAfx.h"
#include "TurbApp.h"
#include "F_Protos.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CTurbApp

BEGIN_MESSAGE_MAP(CTurbApp, CWinApp)
        //{{AFX_MSG_MAP(CTurbApp)
        // NOTE - the ClassWizard will add and remove mapping macros here.
        //    DO NOT EDIT what you see in these blocks of generated code!
        //}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CTurbApp construction

CTurbApp::CTurbApp() = default;

BOOL CTurbApp::InitInstance() {
    SetRegistryKey("Hash, Inc.");
    free((void*)m_pszProfileName);
    m_pszProfileName = _tcsdup("AM Plugins");

    CString string;
    // New Infos here
    m_typeinfoarray.SetArraySize(CELLTYPE_NUMBER);
    m_typeinfoarray[0]->SetMatchName("CELLTYPE_XYZ");
    string.LoadString(IDS_XYZ);
    m_typeinfoarray[0]->SetLocalizedName((LPCTSTR)string);

    m_typeinfoarray[1]->SetMatchName("CELLTYPE_XY");
    string.LoadString(IDS_XY);
    m_typeinfoarray[1]->SetLocalizedName((LPCTSTR)string);

    m_typeinfoarray[2]->SetMatchName("CELLTYPE_SPHEREMAPPED"); // crashes, don't know why yet
    string.LoadString(IDS_SPHEREMAPPED);
    m_typeinfoarray[2]->SetLocalizedName((LPCTSTR)string);

    m_typeinfoarray[3]->SetMatchName("CELLTYPE_CYLINDERMAPPED");
    string.LoadString(IDS_CYLINDERMAPPED);
    m_typeinfoarray[3]->SetLocalizedName((LPCTSTR)string);

    string.LoadString(IDS_CELLTYPE);
    m_celltypeinfo = HTypePropertyInfo::New(string, "CellType", &m_typeinfoarray);
    m_celltypeinfo->SetDefaultIndex(CELLTYPE_XYZ);

    string.LoadString(IDS_OCTAVES);
    m_octavesinfo = HIntPropertyInfo::New(string, "Octaves");
    m_octavesinfo->SetDefaultValue(10);
    m_octavesinfo->SetMinValue(0);
    m_octavesinfo->SetMaxValue(100);

    string.LoadString(IDS_GRIDSIZE);
    m_gridsizeinfo = HIntPropertyInfo::New(string, "Grid Size");
    m_gridsizeinfo->SetDefaultValue(10);
    m_gridsizeinfo->SetMinValue(0);
    m_gridsizeinfo->SetMaxValue(100);

    InitNoiseTable();
    srand(12345);

    return CWinApp::InitInstance();
}

int CTurbApp::ExitInstance() {
    // Delete Infos Here
    delete m_octavesinfo;
    delete m_gridsizeinfo;
    delete m_celltypeinfo;

    return CWinApp::ExitInstance();
}

void CTurbApp::InitNoiseTable() {

    for (int i = MAXX; i-- > 0;)
        for (int j = MAXY; j-- > 0;)
            for (int k = MAXZ; k-- > 0;)
                noiz_tbl[i][j][k] = rand() % 1000;
}

float CTurbApp::Noise(Vector p) const {

    p.x += 15000.0f;
    p.y += 15000.0f;
    p.z += 15000.0f;
    int ix = fast_ftol_unsigned(p.x) % MAXX;
    if (ix < 0)
        ix += MAXX;
    int iy = fast_ftol_unsigned(p.y) % MAXY;
    if (iy < 0)
        iy += MAXY;
    int iz = fast_ftol_unsigned(p.z) % MAXZ;
    if (iz < 0)
        iz += MAXZ;

    const float ox = 1.f - (1.f + hash_math::cos(hash_constants::pi * (p.x - fast_ftol_unsigned(p.x)))) * .5f;
    const float oy = 1.f - (1.f + hash_math::cos(hash_constants::pi * (p.y - fast_ftol_unsigned(p.y)))) * .5f;
    const float oz = 1.f - (1.f + hash_math::cos(hash_constants::pi * (p.z - fast_ftol_unsigned(p.z)))) * .5f;

    const int ixp = ix + 1 >= MAXX ? 0 : ix + 1;
    const int iyp = iy + 1 >= MAXY ? 0 : iy + 1;
    const int izp = iz + 1 >= MAXZ ? 0 : iz + 1;

    const float A = noiz_tbl[ix][iy][iz];
    const float B = noiz_tbl[ixp][iy][iz];
    const float C = noiz_tbl[ixp][iyp][iz];
    const float D = noiz_tbl[ixp][iyp][izp];
    const float E = noiz_tbl[ixp][iy][izp];
    const float F = noiz_tbl[ix][iy][izp];
    const float G = noiz_tbl[ix][iyp][izp];
    const float H = noiz_tbl[ix][iyp][iz];
    float t1 = (1.f - ox) * A + ox * B;
    float t2 = (1.f - ox) * H + ox * C;
    const float t3 = (1.f - ox) * F + ox * E;
    const float t4 = (1.f - ox) * G + ox * D;
    t1 = (1.f - oy) * t1 + oy * t2;
    t2 = (1.f - oy) * t3 + oy * t4;
    t1 = (1.f - oz) * t1 + oz * t2;

    t2 = t1 * .001f;
    if (*(long *)&t2 >= 0)
        return t2;
    return 0.f;
}

float CTurbApp::CellNoise(Vector p) const {
    const float d = 0.001f;
    p.x += 15000.0f;
    p.y += 15000.0f;
    p.z += 15000.0f;
    int ix = fast_ftol_unsigned(p.x) % MAXX;
    if (ix < 0)
        ix += MAXX;
    int iy = fast_ftol_unsigned(p.y) % MAXY;
    if (iy < 0)
        iy += MAXY;
    int iz = fast_ftol_unsigned(p.z) % MAXZ;
    if (iz < 0)
        iz += MAXZ;

    const float t2 = noiz_tbl[ix][iy][iz];

    float t1 = t2 * d;
    if (*(long *)&t1 >= 0)
        return t1;
    return 0.f;
}

/////////////////////////////////////////////////////////////////////////////
// The one and only CTurbApp object

CTurbApp theApp;
