#pragma once
#include "HashMath.h"

#define pulse(a,b,fuzz,x) (smoothstep((a)-(fuzz),(a),(x)) - smoothstep((b)-(fuzz),(b),(x)))
#define repeat(x,freq)    (mod((x) * (freq), 1.))
#define udn(x,lo,hi) (smoothstep(.25, .75, Noise(x)) * ((hi) - (lo)) + (lo))
#define whichtile(x,freq) (fast_floor((x) * (freq)))
#define step(a,x) (((x)<(a)) ? 0 : 1)

inline float mod(float a, const float b) {
    const int n = fast_ftol_unsigned(a / b);

    a -= n * b;
    if (a < 0)
        a += b;
    return a;
}

inline float smoothstep(const float a, const float b, const float x) {
    if (x < a)
        return 0.f;
    if (x >= b)
        return 1.f;
    const float x1 = (x - a) / (b - a);
    return hash_math::sqr(x1) * (3.f - 2.f * x1);
}

inline int Hash(const int k, const int l, const int m) {
    return (k * k * 1312717 + l * l * 976901 + m * m * 897687 + 36 * k + 22 * l + 37 * m + 82168) % 23840231;
}

static float Uniform(const int x, const int y, const int z) {
    return static_cast<float>(Hash(x, y, z) & 0xffff) / 0xffff;
}

static int Poisson(const float lambda, const int n, float x) {
    float p = exp(-lambda);
    int i;
    for (i = 0; i < n; ++i) {
        x -= p;
        if (x < 0)
            return i;
        p = p * lambda / (i + 1.f);
    }
    return i;
}

inline float LocalVoronoi(const int x, const int y, const int z, const Vector *p) {
    float closest = 1;

    //
    // m = number of points in this cell
    //
    const int m = Poisson(4.f, 7, Uniform(x, y, z));
    float cc = 0.f;
    for (int i = 0; i < m; ++i) {
        const float px = x + Uniform(x, y, 2 * i + 1);
        const float py = y + Uniform(x, y, 2 * i + 2);
        const float pz = z + Uniform(x, y, 2 * i + 3);
        const float d2 = (px - p->x) * (px - p->x) + (py - p->y) * (py - p->y) + (pz - p->z) * (pz - p->z);
        if (d2 < closest) {
            closest = d2;
/*
            float closest_x = px;
            float closest_y = py;
            float closest_z = pz;
           */ 
            cc = px;
        }
    }
    if (cc < 0.f)
        cc = 1.f - hash_math::abs(cc);
    if (closest > 0.f)
        return cc;
    return 0.f;
}
